#!python 
#!python
import pycuda.driver as cuda
import pycuda.autoinit
from pycuda.compiler import SourceModule
import pycuda.gpuarray as gpuarray
import numpy as np

# Converting the list into numpy array for faster access and putting it into the GPU for processing... 
start = cuda.Event()
end = cuda.Event()

N = 222341

values = np.random.randn(N)
number_of_blocks=N/1024

# Calculating the (value-max)/max-min computation and storing it in a numpy array. Pre-calculating the maximum and minimum values.

# Space for the Kernel computation..

func_mod = SourceModule("""
// Needed to avoid name mangling so that PyCUDA can
// find the kernel function:
extern "C" {
__global__ void func(float *a, int N, float minval, int denom)
{
int idx = threadIdx.x+threadIdx.y*32+blockIdx.x*blockDim.x;
if (idx < N)
    a[idx] = (a[idx]-minval)/denom;
}
}
""", no_extern_c=1)

func = func_mod.get_function('func')
x = np.asarray(values, np.float32)
x_gpu = gpuarray.to_gpu(x)
h_minval = np.float32(0)
h_denom = np.int32(255)

start.record()
# a function to the GPU to calculate the computation in the GPU.
func(x_gpu.gpudata, np.uint32(N), np.float32(h_minval), np.uint32(h_denom), block=(1024, 1, 1), grid=(number_of_blocks+1,1,1))
end.record() 
end.synchronize()
secs = start.time_till(end)*1e-3

print("SourceModule time")
print("%fs" % (secs))
print('x:       ', x[N-1])
print('Func(x): ', x_gpu.get()[N-1],'Actual: ',(values[N-1]-0)/(h_denom))
x_colors=x_gpu.get()






#!python 
#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Requires PyCuda, PyOpenGL, and Pil
# MAKE SURE YOU HAVE AN UPDATED VERSION OF THESE PACKAGES!!
#
# Ported to PyCUDA by
# Stefano Brilli: stefanobrilli@gmail.com
#
# Copyright 1993-2010 NVIDIA Corporation.  All rights reserved.
#
# This software contains source code provided by NVIDIA Corporation
#
# http://developer.download.nvidia.com/compute/cuda/2_3/sdk/docs/cudasdk_eula.pdf
#
# Please refer to the NVIDIA end user license agreement (EULA) associated
# with this source code for terms and conditions that govern your use of
# this software. Any use, reproduction, disclosure, or distribution of
# this software and related documentation outside the terms of the EULA
# is strictly prohibited.
#

from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *
from OpenGL.GL.ARB.vertex_buffer_object import *
import numpy as np, Image
import sys, time, os
import pycuda.driver as cuda_driver
import pycuda.gl as cuda_gl
import pycuda
#import pycuda.gl.autoinit
from pycuda.compiler import SourceModule

imWidth = 0
imHeight = 0
wWidth = 0
wHeight = 0
wName = "Cuda Edge Detection:"
pixels = None
array = None
texid = 0
pbo_buffer = None
cuda_pbo_resource = None
mode = 0
scale = 1.0

frameCount = 0
fpsCount = 0
fpsLimit = 8
timer = 0.0
ver2011 = False

def copy2D_array_to_device(dst, src, type_sz, width, height):
    copy = cuda_driver.Memcpy2D()
    copy.set_src_array(src)
    copy.set_dst_device(dst)
    copy.height = height
    copy.dst_pitch = copy.src_pitch = copy.width_in_bytes = width*type_sz
    copy(aligned=True)

def computeFPS():
    global frameCount, fpsCount, fpsLimit, timer
    frameCount += 1
    fpsCount += 1
    if fpsCount == fpsLimit:
        ifps = 1.0 /timer
        glutSetWindowTitle("Cuda Edge Detection: %f fps" % ifps)
        fpsCount = 0

def sobelFilter(odata, iw, ih):
    global array, pixels, mode, scale
    if mode == 3:
        # Texture and shared memory with fixed BlockSize
        sm = SourceModule("""
            texture<unsigned char, 2> tex;
            extern __shared__ unsigned char LocalBlock[];
            #define RADIUS 1
            #define BlockWidth 80
            #define SharedPitch 384
            __device__ unsigned char
            ComputeSobel(unsigned char ul, // upper left
                         unsigned char um, // upper middle
                         unsigned char ur, // upper right
                         unsigned char ml, // middle left
                         unsigned char mm, // middle (unused)
                         unsigned char mr, // middle right
                         unsigned char ll, // lower left
                         unsigned char lm, // lower middle
                         unsigned char lr, // lower right
                         float fScale )
            {
                short Horz = ur + 2*mr + lr - ul - 2*ml - ll;
                short Vert = ul + 2*um + ur - ll - 2*lm - lr;
                short Sum = (short) (fScale*(::abs(int(Horz))+::abs(int(Vert))));
                if ( Sum < 0 ) return 0; else if ( Sum > 0xff ) return 0xff;
                return (unsigned char) Sum;
            }

            __global__ void
            SobelShared( int* pSobelOriginal, unsigned short SobelPitch,
                         short w, short h, float fScale )
            {
                short u = 4*blockIdx.x*BlockWidth;
                short v = blockIdx.y*blockDim.y + threadIdx.y;
                short ib;

                int SharedIdx = threadIdx.y * SharedPitch;

                for ( ib = threadIdx.x; ib < BlockWidth+2*RADIUS; ib += blockDim.x ) {
                    LocalBlock[SharedIdx+4*ib+0] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+0), (float) (v-RADIUS) );
                    LocalBlock[SharedIdx+4*ib+1] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+1), (float) (v-RADIUS) );
                    LocalBlock[SharedIdx+4*ib+2] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+2), (float) (v-RADIUS) );
                    LocalBlock[SharedIdx+4*ib+3] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+3), (float) (v-RADIUS) );
                }
                if ( threadIdx.y < RADIUS*2 ) {
                    //
                    // copy trailing RADIUS*2 rows of pixels into shared
                    //
                    SharedIdx = (blockDim.y+threadIdx.y) * SharedPitch;
                    for ( ib = threadIdx.x; ib < BlockWidth+2*RADIUS; ib += blockDim.x ) {
                        LocalBlock[SharedIdx+4*ib+0] = tex2D( tex,
                            (float) (u+4*ib-RADIUS+0), (float) (v+blockDim.y-RADIUS) );
                        LocalBlock[SharedIdx+4*ib+1] = tex2D( tex,
                            (float) (u+4*ib-RADIUS+1), (float) (v+blockDim.y-RADIUS) );
                        LocalBlock[SharedIdx+4*ib+2] = tex2D( tex,
                            (float) (u+4*ib-RADIUS+2), (float) (v+blockDim.y-RADIUS) );
                        LocalBlock[SharedIdx+4*ib+3] = tex2D( tex,
                            (float) (u+4*ib-RADIUS+3), (float) (v+blockDim.y-RADIUS) );
                    }
                }

                __syncthreads();

                u >>= 2;    // index as uchar4 from here
                uchar4 *pSobel = (uchar4 *) (((char *) pSobelOriginal)+v*SobelPitch);
                SharedIdx = threadIdx.y * SharedPitch;

                for ( ib = threadIdx.x; ib < BlockWidth; ib += blockDim.x ) {

                    unsigned char pix00 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+0];
                    unsigned char pix01 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+1];
                    unsigned char pix02 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+2];
                    unsigned char pix10 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+0];
                    unsigned char pix11 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+1];
                    unsigned char pix12 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+2];
                    unsigned char pix20 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+0];
                    unsigned char pix21 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+1];
                    unsigned char pix22 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+2];

                    uchar4 out;

                    out.x = ComputeSobel(pix00, pix01, pix02,
                                         pix10, pix11, pix12,
                                         pix20, pix21, pix22, fScale );

                    pix00 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+3];
                    pix10 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+3];
                    pix20 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+3];
                    out.y = ComputeSobel(pix01, pix02, pix00,
                                         pix11, pix12, pix10,
                                         pix21, pix22, pix20, fScale );

                    pix01 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+4];
                    pix11 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+4];
                    pix21 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+4];
                    out.z = ComputeSobel( pix02, pix00, pix01,
                                          pix12, pix10, pix11,
                                          pix22, pix20, pix21, fScale );

                    pix02 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+5];
                    pix12 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+5];
                    pix22 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+5];
                    out.w = ComputeSobel( pix00, pix01, pix02,
                                          pix10, pix11, pix12,
                                          pix20, pix21, pix22, fScale );
                    if ( u+ib < w/4 && v < h ) {
                        pSobel[u+ib] = out;
                    }
                }

                __syncthreads();
            }
        """)
        cuda_function = sm.get_function("SobelShared")
    elif mode == 2:
        # Texture and shared memory with variable BlockSize
        sm = SourceModule("""
        #define RADIUS 1
        texture<unsigned char, 2> tex;
        extern __shared__ unsigned char LocalBlock[];
        __device__ unsigned char
        ComputeSobel(unsigned char ul, // upper left
                     unsigned char um, // upper middle
                     unsigned char ur, // upper right
                     unsigned char ml, // middle left
                     unsigned char mm, // middle (unused)
                     unsigned char mr, // middle right
                     unsigned char ll, // lower left
                     unsigned char lm, // lower middle
                     unsigned char lr, // lower right
                     float fScale )
        {
            short Horz = ur + 2*mr + lr - ul - 2*ml - ll;
            short Vert = ul + 2*um + ur - ll - 2*lm - lr;
            short Sum = (short) (fScale*(::abs(int(Horz))+::abs(int(Vert))));
            if ( Sum < 0 ) return 0; else if ( Sum > 0xff ) return 0xff;
            return (unsigned char) Sum;
        }

        __global__ void
        SobelShared( int* pSobelOriginal, unsigned short SobelPitch,
                     short BlockWidth, short SharedPitch,
                     short w, short h, float fScale )
        {
            short u = 4*blockIdx.x*BlockWidth;
            short v = blockIdx.y*blockDim.y + threadIdx.y;
            short ib;

            int SharedIdx = threadIdx.y * SharedPitch;

            for ( ib = threadIdx.x; ib < BlockWidth+2*RADIUS; ib += blockDim.x ) {
                LocalBlock[SharedIdx+4*ib+0] = tex2D( tex,
                    (float) (u+4*ib-RADIUS+0), (float) (v-RADIUS) );
                LocalBlock[SharedIdx+4*ib+1] = tex2D( tex,
                    (float) (u+4*ib-RADIUS+1), (float) (v-RADIUS) );
                LocalBlock[SharedIdx+4*ib+2] = tex2D( tex,
                    (float) (u+4*ib-RADIUS+2), (float) (v-RADIUS) );
                LocalBlock[SharedIdx+4*ib+3] = tex2D( tex,
                    (float) (u+4*ib-RADIUS+3), (float) (v-RADIUS) );
            }
            if ( threadIdx.y < RADIUS*2 ) {
                //
                // copy trailing RADIUS*2 rows of pixels into shared
                //
                SharedIdx = (blockDim.y+threadIdx.y) * SharedPitch;
                for ( ib = threadIdx.x; ib < BlockWidth+2*RADIUS; ib += blockDim.x ) {
                    LocalBlock[SharedIdx+4*ib+0] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+0), (float) (v+blockDim.y-RADIUS) );
                    LocalBlock[SharedIdx+4*ib+1] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+1), (float) (v+blockDim.y-RADIUS) );
                    LocalBlock[SharedIdx+4*ib+2] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+2), (float) (v+blockDim.y-RADIUS) );
                    LocalBlock[SharedIdx+4*ib+3] = tex2D( tex,
                        (float) (u+4*ib-RADIUS+3), (float) (v+blockDim.y-RADIUS) );
                }
            }

            __syncthreads();

            u >>= 2;    // index as uchar4 from here
            uchar4 *pSobel = (uchar4 *) (((char *) pSobelOriginal)+v*SobelPitch);
            SharedIdx = threadIdx.y * SharedPitch;

            for ( ib = threadIdx.x; ib < BlockWidth; ib += blockDim.x ) {

                unsigned char pix00 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+0];
                unsigned char pix01 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+1];
                unsigned char pix02 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+2];
                unsigned char pix10 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+0];
                unsigned char pix11 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+1];
                unsigned char pix12 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+2];
                unsigned char pix20 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+0];
                unsigned char pix21 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+1];
                unsigned char pix22 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+2];

                uchar4 out;

                out.x = ComputeSobel(pix00, pix01, pix02,
                                     pix10, pix11, pix12,
                                     pix20, pix21, pix22, fScale );

                pix00 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+3];
                pix10 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+3];
                pix20 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+3];
                out.y = ComputeSobel(pix01, pix02, pix00,
                                     pix11, pix12, pix10,
                                     pix21, pix22, pix20, fScale );

                pix01 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+4];
                pix11 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+4];
                pix21 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+4];
                out.z = ComputeSobel( pix02, pix00, pix01,
                                      pix12, pix10, pix11,
                                      pix22, pix20, pix21, fScale );

                pix02 = LocalBlock[SharedIdx+4*ib+0*SharedPitch+5];
                pix12 = LocalBlock[SharedIdx+4*ib+1*SharedPitch+5];
                pix22 = LocalBlock[SharedIdx+4*ib+2*SharedPitch+5];
                out.w = ComputeSobel( pix00, pix01, pix02,
                                      pix10, pix11, pix12,
                                      pix20, pix21, pix22, fScale );
                if ( u+ib < w/4 && v < h ) {
                    pSobel[u+ib] = out;
                }
            }

            __syncthreads();
        }

        """)
        cuda_function = sm.get_function("SobelShared")
    if mode == 1:
        # Just Texture
        sm = SourceModule("""
        texture<unsigned char, 2> tex;
        __device__ unsigned char ComputeSobel(unsigned char ul, // upper left
                     unsigned char um, // upper middle
                     unsigned char ur, // upper right
                     unsigned char ml, // middle left
                     unsigned char mm, // middle (unused)
                     unsigned char mr, // middle right
                     unsigned char ll, // lower left
                     unsigned char lm, // lower middle
                     unsigned char lr, // lower right
                     float fScale )
        {
            short Horz = ur + 2*mr + lr - ul - 2*ml - ll;
            short Vert = ul + 2*um + ur - ll - 2*lm - lr;
            short Sum = (short) (fScale*(::abs(int(Horz))+::abs(int(Vert))));
            if ( Sum < 0 ) return 0; else if ( Sum > 0xff ) return 0xff;
            return (unsigned char) Sum;
        }
        __global__ void SobelTex( int* pSobelOriginal, unsigned int Pitch,
                  int w, int h, float fScale )
        {
            unsigned char *pSobel =
              (unsigned char *) (((char *) pSobelOriginal)+blockIdx.x*Pitch);
            for ( int i = threadIdx.x; i < w; i += blockDim.x ) {
                unsigned char pix00 = tex2D( tex, (float) i-1, (float) blockIdx.x-1 );
                unsigned char pix01 = tex2D( tex, (float) i+0, (float) blockIdx.x-1 );
                unsigned char pix02 = tex2D( tex, (float) i+1, (float) blockIdx.x-1 );
                unsigned char pix10 = tex2D( tex, (float) i-1, (float) blockIdx.x+0 );
                unsigned char pix11 = tex2D( tex, (float) i+0, (float) blockIdx.x+0 );
                unsigned char pix12 = tex2D( tex, (float) i+1, (float) blockIdx.x+0 );
                unsigned char pix20 = tex2D( tex, (float) i-1, (float) blockIdx.x+1 );
                unsigned char pix21 = tex2D( tex, (float) i+0, (float) blockIdx.x+1 );
                unsigned char pix22 = tex2D( tex, (float) i+1, (float) blockIdx.x+1 );
                pSobel[i] = ComputeSobel(pix00, pix01, pix02,
                                         pix10, pix11, pix12,
                                         pix20, pix21, pix22, fScale );
            }
        }
        """)
        cuda_function = sm.get_function("SobelTex")
    elif mode == 0:
        # Just Copy
        sm = SourceModule("""
        texture<unsigned char, 2> tex;
        __global__ void SobelCopyImage(int* pSobelOriginal, unsigned int Pitch, int w, int h, float fscale )
        {
            unsigned char *pSobel =
              (unsigned char *) (((unsigned char *) pSobelOriginal)+blockIdx.x*Pitch);
            for ( int i = threadIdx.x; i < w; i += blockDim.x ) {
                pSobel[i] = min( max((tex2D( tex, (float)i, (float)blockIdx.x ) * fscale), 0.f), 255.f);
            }
        }
        """)
        cuda_function = sm.get_function("SobelCopyImage")
    texref = sm.get_texref("tex")
    texref.set_array(array)
    texref.set_flags(cuda_driver.TRSA_OVERRIDE_FORMAT)
    if mode == 3:
        # fixed BlockSize Launch
        RADIUS = 1
        threads = (16, 4, 1)
        BlockWidth = 80 # Do not change!
        blocks = (iw/(4*BlockWidth)+(0!=iw%(4*BlockWidth)),
                               ih/threads[1]+(0!=ih%threads[1]) )
        SharedPitch = ~0x3f & (4*(BlockWidth+2*RADIUS)+0x3f);
        sharedMem = SharedPitch*(threads[1]+2*RADIUS);
        iw = iw & ~3
        cuda_function(np.intp(odata), np.uint16(iw), np.int16(iw), np.int16(ih), np.float32(scale), texrefs=[texref],block=threads, grid=blocks, shared=sharedMem)
    elif mode == 2:
        # variable BlockSize launch
        RADIUS = 1
        threads = (16, 4, 1)
        BlockWidth = 80 # Change only with divisible by 16 values!
        blocks = (iw/(4*BlockWidth)+(0!=iw%(4*BlockWidth)),
                               ih/threads[1]+(0!=ih%threads[1]) )
        SharedPitch = ~0x3f & (4*(BlockWidth+2*RADIUS)+0x3f);
        sharedMem = SharedPitch*(threads[1]+2*RADIUS);
        iw = iw & ~3
        cuda_function(np.intp(odata), np.uint16(iw), np.int16(BlockWidth), np.int16(SharedPitch), np.int16(iw), np.int16(ih), np.float32(scale), texrefs=[texref],block=threads, grid=blocks, shared=sharedMem)
    else:
        BlockWidth = 384
        cuda_function(np.intp(odata), np.uint32(iw), np.int32(iw), np.int32(ih), np.float32(scale), texrefs=[texref],block=(BlockWidth,1,1),grid=(ih,1))

def initGL():
    global wWidth, wHeight, wName
    glutInit(sys.argv)
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA)
    glutInitWindowSize(wWidth, wHeight)
    glutCreateWindow(wName)
    import pycuda.gl.autoinit

def loadImage(fn=None):
    global pixels, imWidth, imHeight, wWidth, wHeight
    try:
        im = Image.open(fn) # Open the image
    except IOError:
        print("Usage:", os.path.basename(sys.argv[0]), "[IMAGE=defaultimage.jpg]")
        print("Can't open", fn)
        sys.exit(1)
    imWidth, imHeight = im.size # Window size is set to image size
    wWidth, wHeight = im.size
    im.draft("L", im.size) # L-flag is for Luminance
    pixels = np.fromstring(im.tostring(), dtype=np.uint8) # Got the array
    pixels.resize((imHeight, imWidth)) # Resize to 2d array
    print("Reading image:", fn, "size:", imWidth, "x", imHeight)

def initData(fn=None):
    global pixels, array, pbo_buffer, cuda_pbo_resource, imWidth, imHeight, texid

    # Cuda array initialization
    array = cuda_driver.matrix_to_array(pixels, "C") # C-style instead of Fortran-style: row-major

    pixels.fill(0) # Resetting the array to 0

    pbo_buffer = glGenBuffers(1) # generate 1 buffer reference
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo_buffer) # binding to this buffer
    glBufferData(GL_PIXEL_UNPACK_BUFFER, imWidth*imHeight, pixels, GL_STREAM_DRAW) # Allocate the buffer
    bsize = glGetBufferParameteriv(GL_PIXEL_UNPACK_BUFFER, GL_BUFFER_SIZE) # Check allocated buffer size
    assert(bsize == imWidth*imHeight)
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0) # Unbind

    if ver2011:
        cuda_pbo_resource = pycuda.gl.RegisteredBuffer(int(pbo_buffer), cuda_gl.graphics_map_flags.WRITE_DISCARD)
    else:
        cuda_pbo_resource = cuda_gl.BufferObject(int(pbo_buffer)) # Mapping GLBuffer to cuda_resource


    glGenTextures(1, texid); # generate 1 texture reference
    glBindTexture(GL_TEXTURE_2D, texid); # binding to this texture
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, imWidth, imHeight,  0, GL_LUMINANCE, GL_UNSIGNED_BYTE, None); # Allocate the texture
    glBindTexture(GL_TEXTURE_2D, 0) # Unbind

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1) # 1-byte row alignment
    glPixelStorei(GL_PACK_ALIGNMENT, 1) # 1-byte row alignment


def display():
    global cuda_pbo_resource, pbo_buffer, texid, imWidth, imHeight, timer

    timer = time.time() # Starting timer
    mapping_obj = cuda_pbo_resource.map() # Map the GlBuffer
    if ver2011:
        data, sz = mapping_obj.device_ptr_and_size() # Got the CUDA pointer to GlBuffer
    else:
        data = mapping_obj.device_ptr()
    sobelFilter(data, imWidth, imHeight) # Writing to "data"
    mapping_obj.unmap() # Unmap the GlBuffer

    glClear(GL_COLOR_BUFFER_BIT) # Clear
    glBindTexture(GL_TEXTURE_2D, texid)
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo_buffer)
    # Copyng from buffer to texture
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imWidth, imHeight, GL_LUMINANCE, GL_UNSIGNED_BYTE, None)
    #glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, imWidth, imHeight,  0, GL_LUMINANCE, GL_UNSIGNED_BYTE, None);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0) # Unbind

    glDisable(GL_DEPTH_TEST)
    glEnable(GL_TEXTURE_2D)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)

    glBegin(GL_QUADS)
    glVertex2f(0, 0)
    glTexCoord2f(0, 0)
    glVertex2f(0, 1)
    glTexCoord2f(1, 0)
    glVertex2f(1, 1)
    glTexCoord2f(1, 1)
    glVertex2f(1, 0)
    glTexCoord2f(0, 1)
    glEnd()
    glBindTexture(GL_TEXTURE_2D, 0)
    glutSwapBuffers()
    timer = time.time()-timer
    computeFPS()
    glutPostRedisplay()

def reshape(x, y):
    glViewport(0, 0, x, y)
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, 1, 0, 1, 0, 1)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    glutPostRedisplay()

def keyboard(key, x=0, y=0):
    global mode, scale
    if key=="q":
        sys.exit(0)
    elif key=="I" or key=="i":
        mode = 0
    elif key=="T" or key=="t":
        mode = 1
    elif key=="S" or key=="s":
        mode = 2
    elif key=="D" or key=="d":
        mode = 3
    elif key == "-":
        scale -= 0.1
    elif key == "=":
        scale += 0.1

def idle():
    glutPostRedisplay()

def main(argv):
    fn = "defaultimage.jpg"
    if len(argv) > 1:
        fn = argv[1]

    loadImage(fn) # Loading the image

    initGL()
    initData(fn)
    print("""
    Q: Exit
    I: display image
    T: display Sobel edge detection (computed with tex)
    S: display Sobel edge detection (computed with tex+shared memory)
    D: display Sobel edge detection (computed with tex+shared memory+fixed block size)
    Use the '-' and '=' keys to change the brightness.

    TESTED WITH IMAGE SIZE OF 512x512... just like the original demo.
    Other image sizes may not work
    """)
    glutDisplayFunc(display)
    glutKeyboardFunc(keyboard)
    glutReshapeFunc(reshape)
    glutIdleFunc(idle)
    glutMainLoop();

if __name__ == "__main__":
    if pycuda.VERSION[0] >= 2011:
        ver2011 = True
    main(sys.argv)





#!python 
import pycuda.driver as cuda
import pycuda.autoinit
from pycuda.compiler import SourceModule

mod = SourceModule("""
    #include <stdio.h>

    __global__ void say_hi()
    {
      printf("I am %d.%d\\n", threadIdx.x, threadIdx.y);
    }
    """)

func = mod.get_function("say_hi")
func(block=(4,4,1))




#!python 
# Interactive Mandelbrot Set Accelerated using PyCUDA
# Classical Iteration Method
# Luis Villasenor
# lvillasen@gmail.com
# 2/8/2016
# Licence: GPLv3

# Usage
# Use the left buttom to draw a square to zoom into 

# Point and click with the right buttom to magnify by a factor of 10

# Click with the left button on the right side of the 
# image to randomly change the colormap

# Click with right button on the right side of the image to set the default colormap

# Click on the left side of the image to restart with the full Mandelbrot set

# Press the up/down arrow to increase/decrease the maximum number of iterations

# Press the right/left arrow to increase/decrease the number of pixels

# Type a number from 1-9 to set power index of the iteration formula

# Type 'f' to toggle full-screen mode

# Type 's' to save the image


import pycuda.driver as drv
import pycuda.tools
import pycuda.autoinit
import numpy as np
from pylab import cm as cm
import matplotlib.pyplot as plt
from matplotlib.widgets import RectangleSelector
from matplotlib.patches import Rectangle
from pycuda.compiler import SourceModule
global N,n_block,n_grid,x0,y0,side,L,M,power
L=400;
N=800;n_block=16;n_grid=int(N/16);
N=n_block*n_grid;
x0=-.5;y0=0.
side=3.0
i_cmap=49
power=2
fig = plt.figure(figsize=(12,12))
fig.suptitle('Interactive Mandelbrot Set, Accelerated with PyCUDA')
ax = fig.add_subplot(111)
cmaps=[m for m in cm.datad if not m.endswith("_r")]
N,x0,y0,side,L,power

mod = SourceModule("""
    #include <stdio.h>
    #include <pycuda-complex.hpp>
    #include <math.h>
    typedef   pycuda::complex<double> pyComplex;
__device__ float norma(pyComplex z){
    return norm(z);
}
__global__ void mandelbrot(double x0, double y0,double side, int L,int power,int *M)
{
    int n_x = blockDim.x*gridDim.x;
    int idx = threadIdx.x + blockDim.x*blockIdx.x;
    int idy = threadIdx.y + blockDim.y*blockIdx.y;
    int threadId = idy*n_x+idx;
    float delta = side/n_x;
    pyComplex c( x0-side/2.+delta*idx,y0-side/2.+delta*idy);
    pyComplex z( x0-side/2.+delta*idx,y0-side/2.+delta*idy);
    int h = 0;
    float R = 2.0;
    while( h<L && norma(z)<R){
        z=pow(z,power)+c;
        h+=1;
    }
    M[threadId]=h;
}
""")
M = np.zeros((N,N)).astype(np.int32)
func = mod.get_function("mandelbrot")
func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))
def zoom_on_square(eclick, erelease):
    'eclick and erelease are the press and release events'
    global N,side,x0,y0,myobj,M,power
    x1, y1 = min(eclick.xdata,erelease.xdata),min( eclick.ydata,erelease.ydata)
    x2, y2 = max(eclick.xdata,erelease.xdata),max( eclick.ydata,erelease.ydata)
    #print(" The button you used were: %s %s" % (eclick.button, erelease.button))
    #print ' Nx=%d, Ny=%d, x0=%f, y0=%f'%(x1, y1, x0,y0)
    #print ' Nx=%d, Ny=%d, x0=%f, y0=%f'%(x2, y2, x0,y0)
    x_1=x0+side*(x1-N/2.)/N
    y_1=y0+side*(y1-N/2.)/N
    x_2=x0+side*(x2-N/2.)/N
    y_2=y0+side*(y2-N/2.)/N
    x0=(x_2+x_1)/2.
    y0=(y_2+y_1)/2.
    side=side*(x2-x1+y2-y1)/N/2 # Average of the 2 rectangle sides
    func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))    
    myobj = plt.imshow(M,origin='lower',cmap=cmaps[i_cmap])
    myobj.set_data(M)
    ax.add_patch(Rectangle((1 - .1, 1 - .1), 0.2, 0.2,alpha=1, facecolor='none',fill=None, ))
    ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
    plt.draw()

def key_selector(event):
    global N,side,x0,y0,myobj,M,power,L,i_cmap,n_grid
    #print(' Key pressed.')
    if event.key == 'up':  # Increase max number of iterations
        L=int(L*1.2);
        print(("Maximum number of iterations changed to %d" % L))
        func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))        
        myobj = plt.imshow(M,cmap=cmaps[i_cmap],origin='lower')
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()
    if event.key == 'down':  # Decrease max number of iterations
        L=int(L/1.2);
        print(("Maximum number of iterations changed to %d" % L))
        func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))        
        myobj = plt.imshow(M,cmap=cmaps[i_cmap],origin='lower')
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()
    if event.key == 'right':  # Increase  number of pixels
        N=int(N*1.2);
        n_grid=int(N/16.);
        N=n_block*n_grid;
        M = np.zeros((N,N)).astype(np.int32)
        print(("Number of pixels per dimension changed to %d" % N))
        func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))        
        myobj = plt.imshow(M,cmap=cmaps[i_cmap],origin='lower')
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()
    if event.key == 'left':  # Decrease  number of pixels
        N=int(N/1.2);
        n_grid=int(N/16.);
        N=n_block*n_grid;
        M = np.zeros((N,N)).astype(np.int32)
        print(("Number of pixels per dimension changed to %d" % N))
        func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))        
        myobj = plt.imshow(M,cmap=cmaps[i_cmap],origin='lower')
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()
    if event.key in ['1','2','3','4','5','6','7','8','9'] :  # Decrease  number of pixels
        power=int(event.key)
        if power <10 and power >0 : 
            print(("Power index set to %d" % power))
            i_cmap=49
            side=3.0; x0=-.5;y0=0.;L=200;
            func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))            
            myobj = plt.imshow(M,cmap=cmaps[i_cmap],origin='lower')
            ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
            plt.draw()

key_selector.RS = RectangleSelector(ax, zoom_on_square,
                                       drawtype='box', useblit=True,
                                       button=[1, 3],  # don't use middle button
                                       minspanx=5, minspany=5,
                                       spancoords='pixels')
#                                     interactive=False)
def zoom_on_point(event):
    global N,side,x0,y0,myobj,L,M,i_cmap,power
    #print(" Button pressed: %d" % (event.button))
    #print(' event.x= %f, event.y= %f '%(event.x,event.y))
    if event.button==3 and event.inaxes: # Zoom on clicked point; new side=10% of old side
        x1, y1 = event.xdata, event.ydata
        x0=x0+side*(x1-N/2.)/N
        y0=y0+side*(y1-N/2.)/N
        side=side*.1
        func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))        
        myobj = plt.imshow(M,origin='lower',cmap=cmaps[i_cmap])
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()
    if not event.inaxes and event.x<.3*N : # Click on left side of image to reset to full fractal
        power=2; side=3.0; x0=-.5;y0=0.;i_cmap=49
        func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))        
        myobj = plt.imshow(M,cmap=cmaps[i_cmap],origin='lower')
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()
    if event.button==1 and not event.inaxes and event.x>.7*N : # Left click on right side of image to set a random colormap
        i_cmap_current=i_cmap
        i_cmap=np.random.randint(len(cmaps))
        if i_cmap==i_cmap_current:
            i_cmap-=1
            if i_cmap< 0 : i_cmap=len(cmaps)-1
        #print("color=",i_cmap) 
        myobj = plt.imshow(M,origin='lower',cmap=cmaps[i_cmap])
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw() 
    if event.button==3 and not event.inaxes and event.x>.7*N : # Right click on right side to set default mapolormap 
        i_cmap=49
        myobj = plt.imshow(M,origin='lower',cmap=cmaps[i_cmap])
        ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
        plt.draw()   
fig.canvas.mpl_connect('button_press_event', zoom_on_point)
fig.canvas.mpl_connect('key_press_event', key_selector)
func(np.float64(x0),np.float64(y0),np.float64(side), np.int32(L),np.int32(power),drv.Out(M),block=(n_block,n_block,1),grid=(n_grid,n_grid,1))
ax.set_title('Side=%.2e, x=%.2e, y=%.2e, %s, L=%d'%(side,x0,y0,cmaps[i_cmap],L))
plt.imshow(M,origin='lower',cmap=cmaps[i_cmap])
plt.show()




#!python 
#!/usr/bin/env python
# -*- coding: utf-8 -*-


""" 
Multiples two square matrices together using multiple blocks and shared memory. 
Each thread block is assigned a "tile" of the resulting matrix and is responsible
for generating the elements in that tile.  Each thread in a block computes one element 
of the tile.
"""

import numpy as np
from numpy import linalg as la
from pycuda import driver, compiler, gpuarray, tools

# -- initialize the device
import pycuda.autoinit

kernel_code_template = """
__global__ void MatrixMulKernel(float *A, float *B, float *C)
{

  const uint wA = %(MATRIX_SIZE)s;
  const uint wB = %(MATRIX_SIZE)s;  
  
  // Block index
  const uint bx = blockIdx.x;
  const uint by = blockIdx.y;

  // Thread index
  const uint tx = threadIdx.x;
  const uint ty = threadIdx.y;

  // Index of the first sub-matrix of A processed by the block
  const uint aBegin = wA * %(BLOCK_SIZE)s * by;
  // Index of the last sub-matrix of A processed by the block
  const uint aEnd = aBegin + wA - 1;
  // Step size used to iterate through the sub-matrices of A
  const uint aStep = %(BLOCK_SIZE)s;

  // Index of the first sub-matrix of B processed by the block
  const uint bBegin = %(BLOCK_SIZE)s * bx;
  // Step size used to iterate through the sub-matrices of B
  const uint bStep = %(BLOCK_SIZE)s * wB;

  // The element of the block sub-matrix that is computed
  // by the thread
  float Csub = 0;
  // Loop over all the sub-matrices of A and B required to
  // compute the block sub-matrix
  for (int a = aBegin, b = bBegin;
       a <= aEnd;
       a += aStep, b += bStep) 
    {
      // Shared memory for the sub-matrix of A
      __shared__ float As[%(BLOCK_SIZE)s][%(BLOCK_SIZE)s];
      // Shared memory for the sub-matrix of B
      __shared__ float Bs[%(BLOCK_SIZE)s][%(BLOCK_SIZE)s];

      // Load the matrices from global memory to shared memory
      // each thread loads one element of each matrix
      As[ty][tx] = A[a + wA * ty + tx];
      Bs[ty][tx] = B[b + wB * ty + tx];
      // Synchronize to make sure the matrices are loaded
      __syncthreads();

      // Multiply the two matrices together;
      // each thread computes one element
      // of the block sub-matrix
      for (int k = 0; k < %(BLOCK_SIZE)s; ++k)
        Csub += As[ty][k] * Bs[k][tx];

      // Synchronize to make sure that the preceding
      // computation is done before loading two new
      // sub-matrices of A and B in the next iteration
      __syncthreads();
    }

  // Write the block sub-matrix to global memory;
  // each thread writes one element
  const uint c = wB * %(BLOCK_SIZE)s * by + %(BLOCK_SIZE)s * bx;
  C[c + wB * ty + tx] = Csub;
}
"""

# define the (square) matrix size
MATRIX_SIZE = 4

# define size of blocks and tiles sub-matrix 
# (we assume that the block size is same as tile size)
TILE_SIZE = 2
BLOCK_SIZE = TILE_SIZE

# create two random square matrices
a_cpu = np.random.randn(MATRIX_SIZE, MATRIX_SIZE).astype(np.float32)
b_cpu = np.random.randn(MATRIX_SIZE, MATRIX_SIZE).astype(np.float32)

# compute reference on the CPU to verify GPU computation
c_cpu = np.dot(a_cpu, b_cpu)

# transfer host (CPU) memory to device (GPU) memory 
a_gpu = gpuarray.to_gpu(a_cpu) 
b_gpu = gpuarray.to_gpu(b_cpu)

# create empty gpu array for the result (C = A * B)
c_gpu = gpuarray.empty((MATRIX_SIZE, MATRIX_SIZE), np.float32)

# get the kernel code from the template 
# by specifying the constants MATRIX_SIZE and BLOCK_SIZE
kernel_code = kernel_code_template % { 
    'MATRIX_SIZE': MATRIX_SIZE,
    'BLOCK_SIZE': BLOCK_SIZE,
    }

# compile the kernel code
mod = compiler.SourceModule(kernel_code)

# get the kernel function from the compiled module
matrixmul = mod.get_function("MatrixMulKernel")

# call the kernel on the card
matrixmul(
    # inputs
    a_gpu, b_gpu, 
    # output
    c_gpu, 
    # grid of multiple blocks
    grid = (MATRIX_SIZE // TILE_SIZE, MATRIX_SIZE // TILE_SIZE),
    # block of multiple threads
    block = (TILE_SIZE, TILE_SIZE, 1), 
    )

# print the results
print("-" * 80)
print("Matrix A (GPU):")
print(a_gpu.get())

print("-" * 80)
print("Matrix B (GPU):")
print(b_gpu.get())

print("-" * 80)
print("Matrix C (GPU):")
print(c_gpu.get())

print("-" * 80)
print("CPU-GPU difference:")
print(c_cpu - c_gpu.get())
print("L2 norm:", la.norm(c_cpu - c_gpu.get()))
np.allclose(c_cpu, c_gpu.get())





#!python 
# Derived from a test case by Chris Heuser
# Also see FAQ about PyCUDA and threads.


import pycuda
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
import threading
import numpy

class GPUThread(threading.Thread):
    def __init__(self, number, some_array):
        threading.Thread.__init__(self)

        self.number = number
        self.some_array = some_array

    def run(self):
        self.dev = cuda.Device(self.number)
        self.ctx = self.dev.make_context()

        self.array_gpu = cuda.mem_alloc(some_array.nbytes)
        cuda.memcpy_htod(self.array_gpu, some_array)

        test_kernel(self.array_gpu)
        print("successful exit from thread %d" % self.number)
        self.ctx.pop()

        del self.array_gpu
        del self.ctx

def test_kernel(input_array_gpu):
    mod = SourceModule("""
        __global__ void f(float * out, float * in)
        {
            int idx = threadIdx.x;
            out[idx] = in[idx] + 6;
        }
        """)
    func = mod.get_function("f")

    output_array = numpy.zeros((1,512))
    output_array_gpu = cuda.mem_alloc(output_array.nbytes)

    func(output_array_gpu,
          input_array_gpu,
          block=(512,1,1))
    cuda.memcpy_dtoh(output_array, output_array_gpu)

    return output_array

cuda.init()
some_array = numpy.ones((1,512), dtype=numpy.float32)
num = cuda.Device.count()

gpu_thread_list = []
for i in range(num):
    gpu_thread = GPUThread(i, some_array)
    gpu_thread.start()
    gpu_thread_list.append(gpu_thread)





'''
 * demo_cdpSimplePrint.py
 *
 * Adapted from NVIDIA's "cdpSimplePrint - Simple Print (CUDA Dynamic Parallelism)" sample
 * http://docs.nvidia.com/cuda/cuda-samples/index.html#simple-print--cuda-dynamic-parallelism-
 * http://ecee.colorado.edu/~siewerts/extra/code/example_code_archive/a490dmis_code/CUDA/cuda_work/samples/0_Simple/cdpSimplePrint/cdpSimplePrint.cu
 *
 * From cdpSimplePrint.cu (not sure if this is Ok with NVIDIA's 38-page EULA though...):
 * ---------------------------------------------------------------------------
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 * ---------------------------------------------------------------------------
'''

import sys, os
import pycuda.autoinit
import pycuda.driver as cuda
from pycuda.compiler import DynamicSourceModule

cdpSimplePrint_cu = '''
#include <cstdio>

////////////////////////////////////////////////////////////////////////////////
// Variable on the GPU used to generate unique identifiers of blocks.
////////////////////////////////////////////////////////////////////////////////
__device__ int g_uids = 0;

////////////////////////////////////////////////////////////////////////////////
// Print a simple message to signal the block which is currently executing.
////////////////////////////////////////////////////////////////////////////////
__device__ void print_info( int depth, int thread, int uid, int parent_uid )
{
  if( threadIdx.x == 0 )
  {
    if( depth == 0 )
      printf( "BLOCK %d launched by the host\\n", uid );
    else
    {
      char buffer[32];
      for( int i = 0 ; i < depth ; ++i )
      {
        buffer[3*i+0] = '|';
        buffer[3*i+1] = ' ';
        buffer[3*i+2] = ' ';
      }
      buffer[3*depth] = '\\0';
      printf( "%sBLOCK %d launched by thread %d of block %d\\n", buffer, uid, thread, parent_uid );
    }
  }
  __syncthreads();
}

////////////////////////////////////////////////////////////////////////////////
// The kernel using CUDA dynamic parallelism.
//
// It generates a unique identifier for each block. Prints the information
// about that block. Finally, if the 'max_depth' has not been reached, the
// block launches new blocks directly from the GPU.
////////////////////////////////////////////////////////////////////////////////
__global__ void cdp_kernel( int max_depth, int depth, int thread, int parent_uid )
{
  // We create a unique ID per block. Thread 0 does that and shares the value with the other threads.
  __shared__ int s_uid;
  if( threadIdx.x == 0 ) 
  {
    s_uid = atomicAdd( &g_uids, 1 );
  }
  __syncthreads();

  // We print the ID of the block and information about its parent.
  print_info( depth, thread, s_uid, parent_uid );
  
  // We launch new blocks if we haven't reached the max_depth yet.
  if( ++depth >= max_depth )
  {
    return;
  }
  cdp_kernel<<<gridDim.x, blockDim.x>>>( max_depth, depth, threadIdx.x, s_uid );
}
'''

def main(argv):
    max_depth = 2
    if len(argv) > 1:
        if len(argv) == 2 and argv[1].isdigit() and int(argv[1]) >= 1 and int(argv[1]) <= 8:
            max_depth = int(argv[1])
        else:
            print("Usage: %s <max_depth>\t(where max_depth is a value between 1 and 8)." % argv[0])
            sys.exit(0)

    print("starting Simple Print (CUDA Dynamic Parallelism)")

    mod = DynamicSourceModule(cdpSimplePrint_cu)
    cdp_kernel = mod.get_function('cdp_kernel').prepare('iiii').prepared_call

    print("***************************************************************************")
    print("The CPU launches 2 blocks of 2 threads each. On the device each thread will")
    print("launch 2 blocks of 2 threads each. The GPU we will do that recursively")
    print("until it reaches max_depth=%d\n" % max_depth)
    print("In total 2")
    num_blocks, sum = 2, 2
    for i in range(1, max_depth):
        num_blocks *= 4
        print("+%d" % num_blocks)
        sum += num_blocks
    print("=%d blocks are launched!!! (%d from the GPU)" % (sum, sum-2))
    print("***************************************************************************\n")

    pycuda.autoinit.context.set_limit(cuda.limit.PRINTF_FIFO_SIZE, 11 << 20)  # make this work at depth=8
    pycuda.autoinit.context.set_limit(cuda.limit.DEV_RUNTIME_SYNC_DEPTH, max_depth)

    print("Launching cdp_kernel() with CUDA Dynamic Parallelism:\n")
    cdp_kernel((2,1), (2,1,1), max_depth, 0, 0, -1)

if __name__ == "__main__":
    main(sys.argv)



#!python 
#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" 
Multiplies two square matrices together using a *single* block of threads and 
global memory only. Each thread computes one element of the resulting matrix.
"""

import numpy as np
from pycuda import driver, compiler, gpuarray, tools

# -- initialize the device
import pycuda.autoinit

kernel_code_template = """
__global__ void MatrixMulKernel(float *a, float *b, float *c)
{
    // 2D Thread ID (assuming that only *one* block will be executed)
    int tx = threadIdx.x;
    int ty = threadIdx.y;

    // Pvalue is used to store the element of the matrix
    // that is computed by the thread
    float Pvalue = 0;

    // Each thread loads one row of M and one column of N, 
    //   to produce one element of P.
    for (int k = 0; k < %(MATRIX_SIZE)s; ++k) {
        float Aelement = a[ty * %(MATRIX_SIZE)s + k];
        float Belement = b[k * %(MATRIX_SIZE)s + tx];
        Pvalue += Aelement * Belement;
    }

    // Write the matrix to device memory;
    // each thread writes one element
    c[ty * %(MATRIX_SIZE)s + tx] = Pvalue;
}
"""

# define the (square) matrix size
#  note that we'll only use *one* block of threads here
#  as a consequence this number (squared) can't exceed max_threads,
#  see http://documen.tician.de/pycuda/util.html#pycuda.tools.DeviceData
#  for more information on how to get this number for your device
MATRIX_SIZE = 2

# create two random square matrices
a_cpu = np.random.randn(MATRIX_SIZE, MATRIX_SIZE).astype(np.float32)
b_cpu = np.random.randn(MATRIX_SIZE, MATRIX_SIZE).astype(np.float32)

# compute reference on the CPU to verify GPU computation
c_cpu = np.dot(a_cpu, b_cpu)

# transfer host (CPU) memory to device (GPU) memory 
a_gpu = gpuarray.to_gpu(a_cpu) 
b_gpu = gpuarray.to_gpu(b_cpu)

# create empty gpu array for the result (C = A * B)
c_gpu = gpuarray.empty((MATRIX_SIZE, MATRIX_SIZE), np.float32)

# get the kernel code from the template 
# by specifying the constant MATRIX_SIZE
kernel_code = kernel_code_template % {
    'MATRIX_SIZE': MATRIX_SIZE 
    }

# compile the kernel code 
mod = compiler.SourceModule(kernel_code)

# get the kernel function from the compiled module
matrixmul = mod.get_function("MatrixMulKernel")

# call the kernel on the card
matrixmul(
    # inputs
    a_gpu, b_gpu, 
    # output
    c_gpu, 
    # (only one) block of MATRIX_SIZE x MATRIX_SIZE threads
    block = (MATRIX_SIZE, MATRIX_SIZE, 1),
    )

# print the results
print("-" * 80)
print("Matrix A (GPU):")
print(a_gpu.get())

print("-" * 80)
print("Matrix B (GPU):")
print(b_gpu.get())

print("-" * 80)
print("Matrix C (GPU):")
print(c_gpu.get())

print("-" * 80)
print("CPU-GPU difference:")
print(c_cpu - c_gpu.get())

np.allclose(c_cpu, c_gpu.get())





#!python 
import pycuda.autoinit
import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
import numpy
import numpy.linalg as la




def main_cg():
    from optparse import OptionParser

    parser = OptionParser(
            usage="%prog [options] MATRIX-MARKET-FILE")
    parser.add_option("-s", "--is-symmetric", action="store_true",
            help="Specify that the input matrix is already symmetric")
    options, args = parser.parse_args()

    from pycuda.tools import DeviceMemoryPool, PageLockedMemoryPool
    dev_pool = DeviceMemoryPool()
    pagelocked_pool = PageLockedMemoryPool()

    from scipy.io import mmread
    csr_mat = mmread(args[0]).tocsr().astype(numpy.float32)

    inv_mat_diag = 1/csr_mat.diagonal()

    print("building...")
    from pycuda.sparse.packeted import PacketedSpMV
    spmv = PacketedSpMV(csr_mat, options.is_symmetric, csr_mat.dtype)
    rhs = numpy.random.rand(spmv.shape[0]).astype(spmv.dtype)

    from pycuda.sparse.operator import DiagonalPreconditioner
    if True:
        precon = DiagonalPreconditioner(
                spmv.permute(gpuarray.to_gpu(
                    inv_mat_diag, allocator=dev_pool.allocate)))
    else:
        precon = None

    from pycuda.sparse.cg import solve_pkt_with_cg
    print("start solve")
    for i in range(4):
        start = drv.Event()
        stop = drv.Event()
        start.record()

        rhs_gpu = gpuarray.to_gpu(rhs, dev_pool.allocate)
        res_gpu, it_count, res_count = \
                solve_pkt_with_cg(spmv, rhs_gpu, precon,
                        tol=1e-7 if spmv.dtype == numpy.float64 else 5e-5,
                        pagelocked_allocator=pagelocked_pool.allocate)
        res = res_gpu.get()

        stop.record()
        stop.synchronize()

        elapsed = stop.time_since(start)*1e-3
        est_flops = (csr_mat.nnz*2*(it_count+res_count)
            + csr_mat.shape[0]*(2+2+2+2+2)*it_count)

        if precon is not None:
            est_flops += csr_mat.shape[0] * it_count

        print("residual norm: %g" % (la.norm(csr_mat*res - rhs)/la.norm(rhs)))
        print(("size: %d, elapsed: %g s, %d it, %d residual, it/second: %g, "
                "%g gflops/s" % (
                    csr_mat.shape[0],
                    elapsed, it_count, res_count, it_count/elapsed,
                    est_flops/elapsed/1e9)))

    # TODO: mixed precision
    # TODO: benchmark
    pagelocked_pool.stop_holding()
    dev_pool.stop_holding()





if __name__ == "__main__":
    print("starting...")
    main_cg()






#!python 
# GL interoperability example, by Peter Berrington.
# Draws a rotating teapot, using cuda to invert the RGB value
# each frame

from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *
from OpenGL.GL.ARB.vertex_buffer_object import *
from OpenGL.GL.ARB.pixel_buffer_object import *


import numpy, sys, time
import pycuda.driver as cuda_driver
import pycuda.gl as cuda_gl
from pycuda.compiler import SourceModule

#this is all munged together from the CUDA SDK postprocessGL example.

initial_size = 512,512
current_size = initial_size
animate = True
enable_cuda = True
window = None     # Number of the glut window.
time_of_last_draw = 0.0
time_of_last_titleupdate = 0.0
frames_per_second = 0.0
frame_counter = 0
output_texture = None # pointer to offscreen render target
(source_pbo, dest_pbo, cuda_module, invert,
 pycuda_source_pbo, pycuda_dest_pbo) = [None]*6
heading,pitch,bank = [0.0]*3

def create_PBOs(w,h):
    global source_pbo, dest_pbo, pycuda_source_pbo, pycuda_dest_pbo
    num_texels = w*h
    data = numpy.zeros((num_texels,4),numpy.uint8)
    source_pbo = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, source_pbo)
    glBufferData(GL_ARRAY_BUFFER, data, GL_DYNAMIC_DRAW)
    glBindBuffer(GL_ARRAY_BUFFER, 0)
    pycuda_source_pbo = cuda_gl.BufferObject(int(source_pbo))
    dest_pbo = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, dest_pbo)
    glBufferData(GL_ARRAY_BUFFER, data, GL_DYNAMIC_DRAW)
    glBindBuffer(GL_ARRAY_BUFFER, 0)
    pycuda_dest_pbo = cuda_gl.BufferObject(int(dest_pbo))

def destroy_PBOs():
    global source_pbo, dest_pbo, pycuda_source_pbo, pycuda_dest_pbo
    for pbo in [source_pbo, dest_pbo]:
        glBindBuffer(GL_ARRAY_BUFFER, int(pbo))
        glDeleteBuffers(1, int(pbo))
        glBindBuffer(GL_ARRAY_BUFFER, 0)
    source_pbo,dest_pbo,pycuda_source_pbo,pycuda_dest_pbo = [None]*4

def create_texture(w,h):
    global output_texture
    output_texture = glGenTextures(1)
    glBindTexture(GL_TEXTURE_2D, output_texture)
    # set basic parameters
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
    # buffer data
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                 w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)

def destroy_texture():
    global output_texture
    glDeleteTextures(output_texture)
    output_texture = None

def init_gl():
    Width, Height = current_size
    glClearColor(0.1, 0.1, 0.5, 1.0)
    glDisable(GL_DEPTH_TEST)
    glViewport(0, 0, Width, Height)
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluPerspective(60.0, Width/float(Height), 0.1, 10.0)
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
    glEnable(GL_LIGHT0)
    red   = ( 1.0, 0.1, 0.1, 1.0 )
    white = ( 1.0, 1.0, 1.0, 1.0 )
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  red  )
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, white)
    glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 60.0)

def resize(Width, Height):
    global current_size
    current_size = Width, Height
    glViewport(0, 0, Width, Height)        # Reset The Current Viewport And Perspective Transformation
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluPerspective(60.0, Width/float(Height), 0.1, 10.0)

def do_tick():
    global time_of_last_titleupdate, frame_counter, frames_per_second
    if ((time.clock () * 1000.0) - time_of_last_titleupdate >= 1000.):
        frames_per_second = frame_counter                   # Save The FPS
        frame_counter = 0  # Reset The FPS Counter
        szTitle = "%d FPS" % (frames_per_second )
        glutSetWindowTitle ( szTitle )
        time_of_last_titleupdate = time.clock () * 1000.0
    frame_counter += 1

# The function called whenever a key is pressed. Note the use of Python tuples to pass in: (key, x, y)
def keyPressed(*args):
    global animate, enable_cuda
    # If escape is pressed, kill everything.
    if args[0] == '\033':
        print('Closing..')
        destroy_PBOs()
        destroy_texture()
        exit()
    elif args[0] == 'a':
        print('toggling animation')
        animate = not animate
    elif args[0] == 'e':
        print('toggling cuda')
        enable_cuda = not enable_cuda

def idle():
    global heading, pitch, bank
    if animate:
        heading += 0.2
        pitch   += 0.6
        bank    += 1.0

    glutPostRedisplay()

def display():
    try:
        render_scene()
        if enable_cuda:
            process_image()
            display_image()
        glutSwapBuffers()
    except:
        from traceback import print_exc
        print_exc()
        from os import _exit
        _exit(0)

def process(width, height):
    """ Use PyCuda """
    grid_dimensions   = (width//16,height//16)

    source_mapping = pycuda_source_pbo.map()
    dest_mapping   = pycuda_dest_pbo.map()

    invert.prepared_call(grid_dimensions, (16, 16, 1),
            source_mapping.device_ptr(),
            dest_mapping.device_ptr())

    cuda_driver.Context.synchronize()

    source_mapping.unmap()
    dest_mapping.unmap()

def process_image():
    """ copy image and process using CUDA """
    global pycuda_source_pbo,source_pbo,current_size, dest_pbo
    image_width, image_height = current_size
    assert source_pbo is not None

    # tell cuda we are going to get into these buffers
    pycuda_source_pbo.unregister()

    # activate destination buffer
    glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, int(source_pbo))

    # read data into pbo. note: use BGRA format for optimal performance
    glReadPixels(
             0,                  #start x
             0,                  #start y
             image_width,        #end   x
             image_height,       #end   y
             GL_BGRA,            #format
             GL_UNSIGNED_BYTE,   #output type
             ctypes.c_void_p(0))

    pycuda_source_pbo = cuda_gl.BufferObject(int(source_pbo))

    # run the Cuda kernel
    process(image_width, image_height)
    # blit convolved texture onto the screen
    # download texture from PBO
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, int(dest_pbo))
    glBindTexture(GL_TEXTURE_2D, output_texture)

    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
                    image_width, image_height,
                    GL_BGRA, GL_UNSIGNED_BYTE, ctypes.c_void_p(0))

def display_image():
    """ render a screen sized quad """
    glDisable(GL_DEPTH_TEST)
    glDisable(GL_LIGHTING)
    glEnable(GL_TEXTURE_2D)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
    glMatrixMode(GL_PROJECTION)
    glPushMatrix()
    glLoadIdentity()
    glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0)
    glMatrixMode( GL_MODELVIEW)
    glLoadIdentity()
    glViewport(0, 0, current_size[0], current_size[1])
    glBegin(GL_QUADS)
    glTexCoord2f(0.0, 0.0)
    glVertex3f(-1.0, -1.0, 0.5)
    glTexCoord2f(1.0, 0.0)
    glVertex3f(1.0, -1.0, 0.5)
    glTexCoord2f(1.0, 1.0)
    glVertex3f(1.0, 1.0, 0.5)
    glTexCoord2f(0.0, 1.0)
    glVertex3f(-1.0, 1.0, 0.5)
    glEnd()
    glMatrixMode(GL_PROJECTION)
    glPopMatrix()
    glDisable(GL_TEXTURE_2D)
    glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0)
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0)


def render_scene():
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)# Clear Screen And Depth Buffer
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity ()      # Reset The Modelview Matrix
    glTranslatef(0.0, 0.0, -3.0)
    glRotatef(heading, 1.0, 0.0, 0.0)
    glRotatef(pitch  , 0.0, 1.0, 0.0)
    glRotatef(bank   , 0.0, 0.0, 1.0)
    glViewport(0, 0, current_size[0],current_size[1])
    glEnable(GL_LIGHTING)
    glEnable(GL_DEPTH_TEST)
    glDepthFunc(GL_LESS)
    glutSolidTeapot(1.0)
    do_tick()#just for fps display..
    return True

def main():
    global window, cuda_module, cuda_gl, cuda_driver, invert
    glutInit(sys.argv)
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH)
    glutInitWindowSize(*initial_size)
    glutInitWindowPosition(0, 0)
    window = glutCreateWindow("PyCuda GL Interop Example")
    glutDisplayFunc(display)
    glutIdleFunc(idle)
    glutReshapeFunc(resize)
    glutKeyboardFunc(keyPressed)
    glutSpecialFunc(keyPressed)
    init_gl()

    # create texture for blitting to screen
    create_texture(*initial_size)

    #setup pycuda gl interop
    import pycuda.gl.autoinit
    import pycuda.gl
    cuda_gl = pycuda.gl
    cuda_driver = pycuda.driver

    cuda_module = SourceModule("""
    __global__ void invert(unsigned char *source, unsigned char *dest)
    {
      int block_num        = blockIdx.x + blockIdx.y * gridDim.x;
      int thread_num       = threadIdx.y * blockDim.x + threadIdx.x;
      int threads_in_block = blockDim.x * blockDim.y;
      //Since the image is RGBA we multiply the index 4.
      //We'll only use the first 3 (RGB) channels though
      int idx              = 4 * (threads_in_block * block_num + thread_num);
      dest[idx  ] = 255 - source[idx  ];
      dest[idx+1] = 255 - source[idx+1];
      dest[idx+2] = 255 - source[idx+2];
    }
    """)
    invert = cuda_module.get_function("invert")
    # The argument "PP" indicates that the invert function will take two PBOs as arguments
    invert.prepare("PP")

    # create source and destination pixel buffer objects for processing
    create_PBOs(*initial_size)

    glutMainLoop()

# Print message to console, and kick off the main to get it rolling.
if __name__ == "__main__":
    print("Hit ESC key to quit, 'a' to toggle animation, and 'e' to toggle cuda")
    main()




#!python 
import pycuda.gpuarray as gpuarray
import pycuda.driver as drv
import pycuda.autoinit
import numpy as np

from pycuda.compiler import SourceModule
func_mod = SourceModule("""
template <class T>
__device__ T incr(T x) {
    return (x + 1.0);
}

// Needed to avoid name mangling so that PyCUDA can
// find the kernel function:
extern "C" {
    __global__ void func(float *a, int N)
    {
        int idx = threadIdx.x;
        if (idx < N)
            a[idx] = incr(a[idx]);
    }
}
""", no_extern_c=1)

func = func_mod.get_function('func')

N = 5
x = np.asarray(np.random.rand(N), np.float32)
x_orig = x.copy()
x_gpu = gpuarray.to_gpu(x)

func(x_gpu.gpudata, np.uint32(N), block=(N, 1, 1))
print('x:       ', x)
print('incr(x): ', x_gpu.get())




#!python 
# Exercise 2 from http://webapp.dam.brown.edu/wiki/SciComp/CudaExercises

# Generate an array of random numbers between 0 and 1
# List the indices of those numbers that are greater than a given limit

import pycuda.driver as cuda
import pycuda.autoinit
import pycuda.gpuarray as gpuarray
from pycuda.compiler import SourceModule
import numpy

# Define block size and number of elements per thread
block_size = 512
el_per_thread = 1
multiple_block_size = el_per_thread * block_size

# Create an array of random numbers and set limit
amount = 256*2560
limit = 0.9
assert amount % multiple_block_size == 0
a = numpy.random.rand(amount)
a = a.astype(numpy.float32)
a_gpu = gpuarray.to_gpu(a)

# Initialize array for the selection on device
selec = numpy.zeros_like(a)
selec = selec.astype(numpy.int32)
selec.fill(-1)
selec_gpu = gpuarray.to_gpu(selec)

# Initialize a counter on device
counter_gpu = gpuarray.zeros(1, dtype=numpy.int32)

# Computation on device
mod = SourceModule("""
#define BLOCK_SIZE %(block_size)d
#define EL_PER_THREAD %(el_per_thread)d
// #define USE_LOOPS 1

__global__ void select_them(float *a, int *selec, float limit, int *counter)
{
    __shared__ int selec_smem[EL_PER_THREAD * BLOCK_SIZE];
    __shared__ int counter_smem;
    int *counter_smem_ptr;

    int jump = 16;
    int idx = EL_PER_THREAD * blockIdx.x * BLOCK_SIZE + threadIdx.x +
              (EL_PER_THREAD - 1) * (threadIdx.x / 16) * jump;

    if (threadIdx.x == 1)
    {
        counter_smem_ptr = &counter_smem;
        counter_smem = 0;
    }

    #if (EL_PER_THREAD == 1) && !defined(USE_LOOPS)
        selec_smem[threadIdx.x] = -1;
    #else
        for (int i = 0; i <= EL_PER_THREAD - 1; i++)
            selec_smem[threadIdx.x + i * BLOCK_SIZE] = -1;
    #endif

    __syncthreads();

   // each counting thread writes its index to shared memory

    #if (EL_PER_THREAD == 1) && !defined(USE_LOOPS)
        if (a[idx] >= limit)
            selec_smem[atomicAdd(counter_smem_ptr, 1)] = idx;
    #else
         for (int i = 0; i <= EL_PER_THREAD - 1; i++)
         {
             if (a[idx + i * jump] >= limit)
                 selec_smem[atomicAdd(counter_smem_ptr, 1)] = idx + i * jump;
         }
    #endif

    __syncthreads();

    if (threadIdx.x == 1)
        counter_smem = atomicAdd(counter, counter_smem);

    __syncthreads();

    #if (EL_PER_THREAD == 1) && !defined(USE_LOOPS)
        if (selec_smem[threadIdx.x] >= 0)
            selec[counter_smem + threadIdx.x] = selec_smem[threadIdx.x];
    #else
        for (int i = 0; i <= EL_PER_THREAD - 1; i++)
        {
            if (selec_smem[threadIdx.x + i * jump] >= 0)
                selec[counter_smem + threadIdx.x + i * jump] =
                selec_smem[threadIdx.x + i * jump];
        }
    #endif
}
""" % {"block_size": block_size, "el_per_thread": el_per_thread})

# Prepare function call
func = mod.get_function("select_them")
func.prepare("PPfP")

block = (block_size, 1, 1)
grid = (amount // multiple_block_size, 1)

# Warmup
warmup = 2
for i in range(warmup):
    func.prepared_call(grid, block, a_gpu.gpudata, selec_gpu.gpudata,
                       limit, counter_gpu.gpudata)
    counter_gpu = gpuarray.zeros(1, dtype=numpy.int32)

# Prepare getting the time
start = cuda.Event()
stop = cuda.Event()

# Call function and get time
cuda.Context.synchronize()
start.record()
count = 10
for i in range(count):
    func.prepared_call(grid, block, a_gpu.gpudata, selec_gpu.gpudata,
                       limit, counter_gpu.gpudata)
    counter_gpu = gpuarray.zeros(1, dtype=numpy.int32)
stop.record()

# Copy selection from device to host
selec_gpu.get(selec)

stop.synchronize()

# Evaluate memory bandwidth and verify solution
elems_in_selec = len(numpy.nonzero(selec >= 0))

elapsed_seconds = stop.time_since(start) * 1e-3
print("mem bw:", (a.nbytes + elems_in_selec * 4) / elapsed_seconds / 1e9 * count)

filtered_set = sorted(list(item for item in selec if item != -1))
reference_set = sorted(list(i for i, x in enumerate(a) if x >= limit))
assert filtered_set == reference_set




#!python 
#! /usr/bin/env python
# DO NOT USE THIS AS A BENCHMARK. See
# http://documen.tician.de/pycuda/array.html#pycuda.curandom.rand

import pycuda.autoinit
import pycuda.driver as drv
import pycuda.curandom as curandom
import numpy
import numpy.linalg as la
from pytools import Table




def main():
    import pycuda.gpuarray as gpuarray

    sizes = []
    times = []
    flops = []
    flopsCPU = []
    timesCPU = []

    for power in range(10, 25): # 24
        size = 1<<power
        print(size)
        sizes.append(size)
        a = gpuarray.zeros((size,), dtype=numpy.float32)

        if power > 20:
            count = 100
        else:
            count = 1000

        #start timer
        start = drv.Event()
        end = drv.Event()
        start.record()

        #cuda operation which fills the array with random numbers
        for i in range(count):
            curandom.rand((size, ))

        #stop timer
        end.record()
        end.synchronize()

        #calculate used time
        secs = start.time_till(end)*1e-3

        times.append(secs/count)
        flops.append(size)

        #cpu operations which fills teh array with random data
        a = numpy.array((size,), dtype=numpy.float32)

        #start timer
        start = drv.Event()
        end = drv.Event()
        start.record()

        #cpu operation which fills the array with random data
        for i in range(count):
            numpy.random.rand(size).astype(numpy.float32)

        #stop timer
        end.record()
        end.synchronize()

        #calculate used time
        secs = start.time_till(end)*1e-3

        #add results to variable
        timesCPU.append(secs/count)
        flopsCPU.append(size)


    #calculate pseudo flops
    flops = [f/t for f, t in zip(flops,times)]
    flopsCPU = [f/t for f, t in zip(flopsCPU,timesCPU)]

    #print the data out
    tbl = Table()
    tbl.add_row(("Size", "Time GPU", "Size/Time GPU", "Time CPU","Size/Time CPU","GPU vs CPU speedup"))
    for s, t, f,tCpu,fCpu in zip(sizes, times, flops,timesCPU,flopsCPU):
        tbl.add_row((s,t,f,tCpu,fCpu,f/fCpu))
    print(tbl)


if __name__ == "__main__":
    main()




#!python 

import pycuda
import pycuda.autoinit
import pycuda.gpuarray as gpuarray
import numpy as np

from cgen import *
from codepy.bpl import BoostPythonModule
from codepy.cuda import CudaModule

#Make a host_module, compiled for CPU
host_mod = BoostPythonModule()

#Make a device module, compiled with NVCC
nvcc_mod = CudaModule(host_mod)

#Describe device module code
#NVCC includes
nvcc_includes = [
    'thrust/sort.h',
    'thrust/device_vector.h',
    'cuda.h',
    ]
#Add includes to module
nvcc_mod.add_to_preamble([Include(x) for x in nvcc_includes])

#NVCC function
nvcc_function = FunctionBody(
    FunctionDeclaration(Value('void', 'my_sort'),
                        [Value('CUdeviceptr', 'input_ptr'),
                         Value('int', 'length')]),
    Block([Statement('thrust::device_ptr<float> thrust_ptr((float*)input_ptr)'),
           Statement('thrust::sort(thrust_ptr, thrust_ptr+length)')]))

#Add declaration to nvcc_mod
#Adds declaration to host_mod as well
nvcc_mod.add_function(nvcc_function)

host_includes = [
    'boost/python/extract.hpp',
    ]
#Add host includes to module
host_mod.add_to_preamble([Include(x) for x in host_includes])

host_namespaces = [
    'using namespace boost::python',
    ]

#Add BPL using statement
host_mod.add_to_preamble([Statement(x) for x in host_namespaces])


host_statements = [
    #Extract information from PyCUDA GPUArray
    #Get length
    'tuple shape = extract<tuple>(gpu_array.attr("shape"))',
    'int length = extract<int>(shape[0])',
    #Get data pointer
    'CUdeviceptr ptr = extract<CUdeviceptr>(gpu_array.attr("ptr"))',
    #Call Thrust routine, compiled into the CudaModule
    'my_sort(ptr, length)',
    #Return result
    'return gpu_array',
    ]

host_mod.add_function(
    FunctionBody(
        FunctionDeclaration(Value('object', 'host_entry'),
                            [Value('object', 'gpu_array')]),
        Block([Statement(x) for x in host_statements])))

#Print out generated code, to see what we're actually compiling
print("---------------------- Host code ----------------------")
print((host_mod.generate()))
print("--------------------- Device code ---------------------")
print((nvcc_mod.generate()))
print("-------------------------------------------------------")



#Compile modules
import codepy.jit, codepy.toolchain
gcc_toolchain = codepy.toolchain.guess_toolchain()
nvcc_toolchain = codepy.toolchain.guess_nvcc_toolchain()

module = nvcc_mod.compile(gcc_toolchain, nvcc_toolchain, debug=True)



length = 100
a = np.array(np.random.rand(length), dtype=np.float32)
print("---------------------- Unsorted -----------------------")
print(a)
b = gpuarray.to_gpu(a)
# Call Thrust!!
c = module.host_entry(b)
print("----------------------- Sorted ------------------------")
print(c.get())
print("-------------------------------------------------------")




#!python 
import pycuda.gpuarray as gpuarray
import pycuda.driver as cuda
import pycuda.tools as tools
import pycuda.autoinit
import numpy, random, time
from pycuda.curandom import rand as curand
from pycuda.elementwise import ElementwiseKernel as Elementwise

x = 50
y = 50
z = 2
width = 100
height = 100
depth = 100

def main():
    """
    Computes a set of distances from a given point in a search space in parallel on a GPU.
    """

    # Create an empty array to hold our points.
    n = gpuarray.zeros(shape=(x, y, z),
                    dtype=gpuarray.vec.float3)

    # Populate the array with randomized points from the search space.
    for k in range(z):
        for j in range(y):
            for i in range(x):
                n[i, j, k] = gpuarray.vec.make_float3(random.uniform(-width, width),
                                                    random.uniform(-height, height),
                                                    random.uniform(-depth, depth))

    # Declare our elementwise CUDA kernel.
    mod = Elementwise(
        arguments="float3 pt, float3 *ns, float *rs",
        operation="rs[i] = sqrt(pow(pt.x-ns[i].x,2)+pow(pt.y-ns[i].y,2)+pow(pt.z-ns[i].z,2))",
        name="euclidean_distance",
        preamble="#include <math.h>"
    )

    # Declare an empty results array.
    r = gpuarray.zeros(shape=(50, 50, 2), dtype=numpy.float32)
    start = cuda.Event()
    end = cuda.Event()
    start.record()
    # Call the kernel with a randomize point from the search space.
    mod(gpuarray.vec.make_float3(random.uniform(-width, width),
                                 random.uniform(-height, height),
                                 random.uniform(-width, width)), n, r)
    end.record()
    end.synchronize()
    print((start.time_till(end)))
    print(r)

if __name__ == '__main__':
    main()





#!python 
"""
3D display of Light Field images.
Example images can be download from:
http://graphics.stanford.edu/software/LFDisplay/LFDisplay-samples.zip

Use:
>> python LFview.py <path to light-field image>

Prerequisites:
- Python 2.7
- Enthought Tool Suite (ETS)
- PIL
- Jinja

Author: Amit Aides. amitibo at technion . ac . il
"""


from enthought.traits.api import HasTraits, Range, on_trait_change
from enthought.traits.ui.api import View, Item
from enthought.chaco.api import Plot, ArrayPlotData, gray
from enthought.enable.component_editor import ComponentEditor

import numpy as np
import Image
import argparse
import os.path
import math

import pycuda.driver as cuda
import pycuda.compiler
import pycuda.autoinit

from jinja2 import Template


_kernel_tpl = Template("""
{% if NCHANNELS == 3 %}
texture<float4, 2> tex;
{% else %}
texture<float, 2> tex;
{% endif %}

__global__ void LFview_kernel(
    float x_angle,
    float y_angle,
    unsigned char* data
    )
{
    //
    // calculate pixel idx
    //
    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;

    //
    // We might be outside the reachable pixels. Don't do anything
    //
    if( (x >= {{newiw}}) || (y >= {{newih}}) )
        return;

    //
    // calculate offset into destination array
    //
    unsigned int didx = (y * {{newiw}} + x) * {{NCHANNELS}};
    
    //
    // calculate offset into source array (be aware of rotation and scaling)
    //
    float sx = {{x_start}} + tan(x_angle)*{{x_ratio}} + {{x_step}}*x;
    float sy = {{y_start}} + tan(y_angle)*{{y_ratio}} + {{y_step}}*y;

    if( (sx < 0) || (sx >= {{oldiw}}) || (sy < 0) || (sy >= {{oldih}}) ) {

        {% for channel in range(NCHANNELS) %}
        data[didx+{{channel}}] = 0;
        {% endfor %}

        return;
    }

    {% if NCHANNELS == 3 %}
    float4 texval = tex2D(tex, sx, sy);
    data[didx] = texval.x;
    data[didx+1] = texval.y;
    data[didx+2] = texval.z;
    {% else %}
    data[didx] = tex2D(tex, sx, sy);
    {% endif %}
}
""")


def ceil(x):
    return int(x + 0.5)


class LFapplication(HasTraits):

    traits_view = View(
        Item('LF_img', editor=ComponentEditor(), show_label=False),
        Item('X_angle', label='Angle in the X axis'),
        Item('Y_angle', label='Angle in the Y axis'),
        resizable = True,
        title="LF Image"
        )

    def __init__(self, img_path):
        super().__init__()

        #
        # Load image data
        #
        base_path = os.path.splitext(img_path)[0]
        lenslet_path = base_path + '-lenslet.txt'
        optics_path = base_path + '-optics.txt'

        with open(lenslet_path) as f:
            tmp = eval(f.readline())
            x_offset, y_offset, right_dx, right_dy, down_dx, down_dy = \
              np.array(tmp, dtype=np.float32)

        with open(optics_path) as f:
            for line in f:
                name, val = line.strip().split()
                try:
                    setattr(self, name, np.float32(val))
                except:
                    pass

        max_angle = math.atan(self.pitch/2/self.flen)

        #
        # Prepare image
        #
        im_pil = Image.open(img_path)
        if im_pil.mode == 'RGB':
            self.NCHANNELS = 3
            w, h = im_pil.size
            im = np.zeros((h, w, 4), dtype=np.float32)
            im[:, :, :3] = np.array(im_pil).astype(np.float32)
            self.LF_dim = (ceil(h/down_dy), ceil(w/right_dx), 3)
        else:
            self.NCHANNELS = 1
            im = np.array(im_pil.getdata()).reshape(im_pil.size[::-1]).astype(np.float32)
            h, w = im.shape
            self.LF_dim = (ceil(h/down_dy), ceil(w/right_dx))

        x_start = x_offset - int(x_offset / right_dx) * right_dx
        y_start = y_offset - int(y_offset / down_dy) * down_dy
        x_ratio = self.flen * right_dx / self.pitch
        y_ratio = self.flen * down_dy / self.pitch

        #
        # Generate the cuda kernel
        #
        mod_LFview = pycuda.compiler.SourceModule(
            _kernel_tpl.render(
                newiw=self.LF_dim[1],
                newih=self.LF_dim[0],
                oldiw=w,
                oldih=h,
                x_start=x_start,
                y_start=y_start,
                x_ratio=x_ratio,
                y_ratio=y_ratio,
                x_step=right_dx,
                y_step=down_dy,
                NCHANNELS=self.NCHANNELS
                )
            )
        
        self.LFview_func = mod_LFview.get_function("LFview_kernel")
        self.texref = mod_LFview.get_texref("tex")
        
        #
        # Now generate the cuda texture
        #
        if self.NCHANNELS == 3:
            cuda.bind_array_to_texref(
                cuda.make_multichannel_2d_array(im, order="C"),
                self.texref
                )
        else:
            cuda.matrix_to_texref(im, self.texref, order="C")
            
        #
        # We could set the next if we wanted to address the image
        # in normalized coordinates ( 0 <= coordinate < 1.)
        # texref.set_flags(cuda.TRSF_NORMALIZED_COORDINATES)
        #
        self.texref.set_filter_mode(cuda.filter_mode.LINEAR)

        #
        # Prepare the traits
        #
        self.add_trait('X_angle', Range(-max_angle, max_angle, 0.0))
        self.add_trait('Y_angle', Range(-max_angle, max_angle, 0.0))
        
        self.plotdata = ArrayPlotData(LF_img=self.sampleLF())
        self.LF_img = Plot(self.plotdata)
        if self.NCHANNELS == 3:
            self.LF_img.img_plot("LF_img")
        else:
            self.LF_img.img_plot("LF_img", colormap=gray)

    def sampleLF(self):
        #
        # Get the output image
        #
        output = np.zeros(self.LF_dim, dtype=np.uint8)
        
        #
        # Calculate the gridsize. This is entirely given by the size of our image. 
        #
        blocks = (16, 16, 1)
        gridx = ceil(self.LF_dim[1]/blocks[1])
        gridy = ceil(self.LF_dim[0]/blocks[0])
        grid = (gridx, gridy)

        #
        # Call the kernel
        #
        self.LFview_func(
            np.float32(self.X_angle),
            np.float32(self.Y_angle),
            cuda.Out(output),
            texrefs=[self.texref],
            block=blocks,
            grid=grid
            )

        return output

    @on_trait_change('X_angle, Y_angle')        
    def updateImge(self):
        self.plotdata.set_data('LF_img', self.sampleLF())
        
        
def main(img_path):
    """Main function"""

    app = LFapplication(img_path)
    app.configure_traits()
    
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='View an LF image')
    parser.add_argument('img_path', type=str, help='Path to LF image')
    args = parser.parse_args()

    main(args.img_path)




#! /usr/bin/env python
import pycuda.driver as drv
import pycuda.autoinit
import numpy
import numpy.linalg as la


def main():
    import pycuda.gpuarray as gpuarray

    sizes = []
    times_gpu = []
    flops_gpu = []
    flops_cpu = []
    times_cpu = []

    from pycuda.tools import bitlog2

    max_power = bitlog2(drv.mem_get_info()[0]) - 2
    # they're floats, i.e. 4 bytes each
    for power in range(10, max_power):
        size = 1 << power
        print(size)
        sizes.append(size)
        a = gpuarray.zeros((size,), dtype=numpy.float32)
        b = gpuarray.zeros((size,), dtype=numpy.float32)
        b.fill(1)

        if power > 20:
            count = 100
        else:
            count = 1000

        # gpu -----------------------------------------------------------------
        start = drv.Event()
        end = drv.Event()
        start.record()

        for i in range(count):
            a + b

        end.record()
        end.synchronize()

        secs = start.time_till(end) * 1e-3

        times_gpu.append(secs / count)
        flops_gpu.append(size)
        del a
        del b

        # cpu -----------------------------------------------------------------
        a_cpu = numpy.random.randn(size).astype(numpy.float32)
        b_cpu = numpy.random.randn(size).astype(numpy.float32)

        # start timer
        from time import time

        start = time()
        for i in range(count):
            a_cpu + b_cpu
        secs = time() - start

        times_cpu.append(secs / count)
        flops_cpu.append(size)

    # calculate pseudo flops
    flops_gpu = [f / t for f, t in zip(flops_gpu, times_gpu)]
    flops_cpu = [f / t for f, t in zip(flops_cpu, times_cpu)]

    from pytools import Table

    tbl = Table()
    tbl.add_row(
        (
            "Size",
            "Time GPU",
            "Size/Time GPU",
            "Time CPU",
            "Size/Time CPU",
            "GPU vs CPU speedup",
        )
    )
    for s, t, f, t_cpu, f_cpu in zip(sizes, times_gpu, flops_gpu, times_cpu, flops_cpu):
        tbl.add_row((s, t, f, t_cpu, f_cpu, f / f_cpu))
    print(tbl)


if __name__ == "__main__":
    main()



#!python 

__author__ = 'ashwin'

import pycuda.driver as drv
import pycuda.tools
import pycuda.autoinit
from pycuda.compiler import SourceModule
import numpy as np
import scipy.misc as scm
import matplotlib.pyplot as p

mod = SourceModule \
    (
        """
#include<stdio.h>
#define INDEX(a, b) a*256+b

__global__ void rgb2gray(float *dest,float *r_img, float *g_img, float *b_img)
{

unsigned int idx = threadIdx.x+(blockIdx.x*(blockDim.x*blockDim.y));

  unsigned int a = idx/256;
  unsigned int b = idx%256;

dest[INDEX(a, b)] = (0.299*r_img[INDEX(a, b)]+0.587*g_img[INDEX(a, b)]+0.114*b_img[INDEX(a, b)]);


}

"""
    )

a = scm.imread('Lenna.png').astype(np.float32)
print(a)
r_img = a[:, :, 0].reshape(65536, order='F')
g_img = a[:, :, 1].reshape(65536, order='F')
b_img = a[:, :, 2].reshape(65536, order='F')
dest=r_img
print(dest)
rgb2gray = mod.get_function("rgb2gray")
rgb2gray(drv.Out(dest), drv.In(r_img), drv.In(g_img),drv.In(b_img),block=(1024, 1, 1), grid=(64, 1, 1))

dest=np.reshape(dest,(256,256), order='F')


p.imshow(dest)
p.show()






#!python 
#!/usr/bin/env python -tt
# encoding: utf-8
#
# Created by Holger Rapp on 2009-03-11.
# HolgerRapp@gmx.net
#

import pycuda.driver as cuda
import pycuda.compiler
import pycuda.autoinit
import numpy
from math import pi,cos,sin

_rotation_kernel_source = """
texture<float, 2> tex;

__global__ void copy_texture_kernel(
    const float resize_val, 
    const float alpha, 
    unsigned short oldiw, unsigned short oldih,
    unsigned short newiw, unsigned short newih,
    unsigned char* data) {

        // calculate pixel idx
        unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
        unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;
        
        // We might be outside the reachable pixels. Don't do anything
        if( (x >= newiw) || (y >= newih) )
            return;
        
        // calculate offset into destination array
        unsigned int didx = y * newiw + x;
        
        // calculate offset into source array (be aware of rotation and scaling)
        float xmiddle = (x-newiw/2.) / resize_val;
        float ymiddle = (y-newih/2.) / resize_val;
        float sx = ( xmiddle*cos(alpha)+ymiddle*sin(alpha) + oldiw/2.) ;
        float sy = ( -xmiddle*sin(alpha)+ymiddle*cos(alpha) + oldih/2.);
        
        if( (sx < 0) || (sx >= oldiw) || (sy < 0) || (sy >= oldih) ) { 
            data[didx] = 255; 
            return;
        }

        data[didx] = tex2D(tex, sx, sy);
    }
"""
mod_copy_texture=pycuda.compiler.SourceModule( _rotation_kernel_source )

copy_texture_func = mod_copy_texture.get_function("copy_texture_kernel")
texref = mod_copy_texture.get_texref("tex")

def rotate_image( a, resize = 1.5, angle = 20., interpolation = "linear", blocks = (16,16,1)  ):
    """
    Rotates the array. The new array has the new size and centers the
    picture in the middle.
    
    a             - array (2-dim)
    resize        - new_image w/old_image w
    angle         - degrees to rotate the image
    interpolation - "linear" or None
    blocks        - given to the kernel when run


    returns: a new array with dtype=uint8 containing the rotated image
    """
    angle = angle/180. *pi
    
    # Convert this image to float. Unsigned int texture gave 
    # strange results for me. This conversion is slow though :(
    a = a.astype("float32")

    # Calculate the dimensions of the new image
    calc_x = lambda x_y: (x_y[0]*a.shape[1]/2.*cos(angle)-x_y[1]*a.shape[0]/2.*sin(angle))
    calc_y = lambda x_y1: (x_y1[0]*a.shape[1]/2.*sin(angle)+x_y1[1]*a.shape[0]/2.*cos(angle))

    xs = [ calc_x(p) for p in [ (-1.,-1.),(1.,-1.),(1.,1.),(-1.,1.) ] ]
    ys = [ calc_y(p) for p in [ (-1.,-1.),(1.,-1.),(1.,1.),(-1.,1.) ] ]

    new_image_dim = (
        int(numpy.ceil(max(ys)-min(ys))*resize),
        int(numpy.ceil(max(xs)-min(xs))*resize),
    )
    
    # Now generate the cuda texture
    cuda.matrix_to_texref(a, texref, order="C")
    
    # We could set the next if we wanted to address the image
    # in normalized coordinates ( 0 <= coordinate < 1.)
    # texref.set_flags(cuda.TRSF_NORMALIZED_COORDINATES)
    if interpolation == "linear":
        texref.set_filter_mode(cuda.filter_mode.LINEAR)

    # Calculate the gridsize. This is entirely given by the size of our image. 
    gridx = new_image_dim[0]/blocks[0] if \
            new_image_dim[0]%blocks[0]==1 else new_image_dim[0]/blocks[0] +1
    gridy = new_image_dim[1]/blocks[1] if \
            new_image_dim[1]%blocks[1]==0 else new_image_dim[1]/blocks[1] +1

    # Get the output image
    output = numpy.zeros(new_image_dim,dtype="uint8")
    
    # Call the kernel
    copy_texture_func(
        numpy.float32(resize), numpy.float32(angle),
        numpy.uint16(a.shape[1]), numpy.uint16(a.shape[0]),
        numpy.uint16(new_image_dim[1]), numpy.uint16(new_image_dim[0]),
            cuda.Out(output),texrefs=[texref],block=blocks,grid=(gridx,gridy))
    
    return output

if __name__ == '__main__':
    import Image
    import sys
    
    def main( ):
        if len(sys.argv) != 2:
            print("You should really read the source...\n\nUsage: rotate.py <Imagename>\n")
            sys.exit(-1)

        # Open, convert to grayscale, convert to numpy array
        img = Image.open(sys.argv[1]).convert("L")
        i = numpy.fromstring(img.tostring(),dtype="uint8").reshape(img.size[1],img.size[0])
        
        # Rotate & convert back to PIL Image
        irot = rotate_image(i)
        rotimg = Image.fromarray(irot,mode="L")

        # Save and display
        rotimg.save("rotated.png")
        rotimg.show()
    
    main()






# Copyright 2008-2021 Andreas Kloeckner
# Copyright 2021 NVIDIA Corporation

from numba import cuda

import pycuda.driver as pycuda
# We use autoprimaryctx instead of autoinit because Numba can only operate on a
# primary context
import pycuda.autoprimaryctx  # noqa
import pycuda.gpuarray as gpuarray

import numpy


# Create a PyCUDA gpuarray
a_gpu = gpuarray.to_gpu(numpy.random.randn(4, 4).astype(numpy.float32))
print("original array:")
print(a_gpu)


# A standard Numba kernel that doubles its input array
@cuda.jit
def double(x):
    i, j = cuda.grid(2)

    if i < x.shape[0] and j < x.shape[1]:
        x[i, j] *= 2


# Call the Numba kernel on the PyCUDA gpuarray, using the CUDA Array Interface
# transparently
double[(4, 4), (1, 1)](a_gpu)
print("doubled with numba:")
print(a_gpu)



#!python 
# Conway's Game of Life Accelerated with PyCUDA
# Luis Villasenor
# lvillasen@gmail.com
# 3/26/2016
# Licence: GPLv3
# Usage: python GameOfLife.py n n_iter
# where n is the board size and n_iter the number of iterations
import pycuda.driver as cuda
import pycuda.tools
import pycuda.autoinit
import pycuda.gpuarray as gpuarray
from pycuda.compiler import SourceModule
import sys
import numpy as np
from pylab import cm as cm
import matplotlib.pyplot as plt
n=int(sys.argv[1])
n_iter=int(sys.argv[2])
n_block=16
n_grid=int(n/n_block);
n=n_block*n_grid;
def random_init(n):
    #np.random.seed(100)
    M=np.zeros((n,n)).astype(np.int32)
    for i in range(n):
        for j in range(n):
            M[j,i]=np.int32(np.random.randint(2))
    return M
mod = SourceModule("""
__global__ void step(int *C, int *M)
{
  int count;
  int n_x = blockDim.x*gridDim.x;
  int i = threadIdx.x + blockDim.x*blockIdx.x;
  int j = threadIdx.y + blockDim.y*blockIdx.y;
  int threadId = j*n_x+i;
  int i_left; int i_right; int j_down; int j_up;
  if(i==0) {i_left=n_x-1;} else {i_left=i-1;}
  if(i==n_x-1) {i_right=0;} else {i_right=i+1;}
  if(j==0) {j_down=n_x-1;} else {j_down=j-1;}
  if(j==n_x-1) {j_up=0;} else {j_up=j+1;}
  count = C[j*n_x+i_left] + C[j_down*n_x+i]
    + C[j*n_x+i_right] + C[j_up*n_x+i] + C[j_up*n_x+i_left]
    + C[j_down*n_x+i_right] + C[j_down*n_x+i_left]
    + C[j_up*n_x+i_right];

// Modify matrix M according to the rules B3/S23:
//A cell is "Born" if it has exactly 3 neighbours,
//A cell "Survives" if it has 2 or 3 living neighbours; it dies otherwise.
  if(count < 2 || count > 3) M[threadId] = 0; // cell dies
  if(count == 2) M[threadId] = C[threadId];// cell stays the same
  if(count == 3) M[threadId] = 1; // cell either stays alive, or is born
}
""")
func = mod.get_function("step")
C=random_init(n)
M = np.empty_like(C)
C_gpu = gpuarray.to_gpu( C )
M_gpu = gpuarray.to_gpu( M )
for k in range(n_iter):
  func(C_gpu,M_gpu,block=(n_block,n_block,1),grid=(n_grid,n_grid,1))
  C_gpu, M_gpu = M_gpu, C_gpu
print(("%d live cells after %d iterations" %(np.sum(C_gpu.get()),n_iter)))
fig = plt.figure(figsize=(12,12))
ax = fig.add_subplot(111)
fig.suptitle("Conway's Game of Life Accelerated with PyCUDA")
ax.set_title('Number of Iterations = %d'%(n_iter))
myobj =plt.imshow(C_gpu.get(),origin='lower',cmap='Greys',  interpolation='nearest',vmin=0, vmax=1)
plt.pause(.01)
plt.draw()
m=n_iter
while True:
    m+=1
    func(C_gpu,M_gpu,block=(n_block,n_block,1),grid=(n_grid,n_grid,1))
    C_gpu, M_gpu = M_gpu, C_gpu
    myobj.set_data(C_gpu.get())
    ax.set_title('Number of Iterations = %d'%(m))
    plt.pause(.01)
    plt.draw()




#!python 
# simple module to show the plotting of random data

import pycuda.autoinit
import pycuda.curandom as curandom

size = 1000
a = curandom.rand((size,)).get()

from matplotlib.pylab import *
subplot(211)
plot(a)
grid(True)
ylabel('plot - gpu')

subplot(212)
hist(a, 100)
grid(True)
ylabel('histogram - gpu')

#and save it
savefig('plot-random-data')





#!python 
import numpy
import scipy.misc
import numpy.fft as nfft
import multiprocessing

from pyfft.cuda import Plan
from pycuda.tools import make_default_context
import pycuda.tools as pytools
import pycuda.gpuarray as garray
import pycuda.driver as drv


class GPUMulti(multiprocessing.Process):
    def __init__(self, number, input_cpu, output_cpu):
        multiprocessing.Process.__init__(self)
        self.number = number
        self.input_cpu = input_cpu
        self.output_cpu = output_cpu

    def run(self):
        drv.init()
        a0=numpy.zeros((p,),dtype=numpy.complex64)
        self.dev = drv.Device(self.number)
        self.ctx = self.dev.make_context()
#TO VERIFY WHETHER ALL THE MEMORY IS FREED BEFORE NEXT ALLOCATION (THIS DOES NOT HAPPEN IN MULTITHREADING)
        print(drv.mem_get_info())
        self.gpu_a = garray.empty((self.input_cpu.size,), dtype=numpy.complex64)
        self.gpu_b = garray.zeros_like(self.gpu_a)
        self.gpu_a = garray.to_gpu(self.input_cpu)
        plan = Plan(a0.shape,context=self.ctx)
        plan.execute(self.gpu_a, self.gpu_b, batch=p/m)
        self.temp = self.gpu_b.get()
        self.output_cpu.put(self.temp)
        self.output_cpu.close()
        self.ctx.pop()
        del self.gpu_
        del self.gpu_b
        del self.ctx

        print("till the end %d" %self.number)


p = 8192; # INPUT IMAGE SIZE (8192 * 8192)
m = 4     # TO DIVIDE THE INPUT IMAGE INTO 4* (2048 * 8192) SIZED IMAGES (Depends on the total memory of your GPU)
trans = 2 # FOR TRANSPOSE-SPLIT (TS) ALGORITHM WHICH loops 2 times


#INPUT IMAGE (GENERATE A 2d SINE WAVE PATTERN)
p_n = 8000 # No. OF PERIODS OF SINE WAVES
x=numpy.arange(0,p_n,float(p_n)/float(p))
a_i = 128 + 128 * numpy.sin(2*numpy.pi*x)
a2 = numpy.zeros([p,p],dtype=numpy.complex64)
a2[::]=a_i
scipy.misc.imsave("sine.bmp",numpy.absolute(a2)) #TEST THE GENERATION OF INPUT IMAGE

#INITIALISE THE VARIABLES
a2_1 = numpy.zeros([m,p*p/m],dtype = numpy.complex64) #INPUT TO THE GPU (1d ARRAY)
#VERY IMPORTANT
output_cpu  = multiprocessing.Queue() #STORE RESULT IN GPU (MULTIPROCESSING DOES NOT ALLOW SHARING AND HENCE THIS IS NEEDED FOR COMMUNICATION OF DATA)

b2pa = numpy.zeros([p/m,p,m],dtype = numpy.complex64) #OUTPUT FROM GPU
b2_a = numpy.zeros([p,p],dtype = numpy.complex64)     #RESHAPED (8192*8192) OUTPUT

#NOW WE ARE READY TO KICK START THE GPU

# THE NO OF GPU'S PRESENT (CHANGE ACCORDING TO THE No.OF GPUS YOU HAVE)
num = 2 # I KNOW THIS IS A BAD PRACTISE, BUT I COUNDN'T FIND ANY OTHER WAY(INIT CANNOT BE USED HERE)

#THE TRANSPOSE-SPLIT ALGORITHM FOR FFT
for t in range (0,trans):
    for i in range (m):
        a2_1[i,:] = a2[i*p/m:(i+1)*p/m,:].flatten()#DIVIDE AND RESHAPE THE INPUT IMAGE INTO 1D ARRAY

    for j in range (m/num):
        gpu_multi_list = []

#CREATE AND START THE MULTIPROCESS
        for i in range (num):
            gpu_multi = GPUMulti(i,a2_1[i+j*num,:],output_cpu) #FEED THE DATA INTO THE GPU
            gpu_multi_list.append(gpu_multi)
            gpu_multi.start()#THERE YOU GO

#COLLECT THE OUTPUT FROM THE RUNNING MULTIPROCESS AND RESHAPE
        for gpu_pro in gpu_multi_list:
            temp_b2_1 = output_cpu.get(gpu_pro)
            b2pa[:,:,gpu_pro.number+j*num] = numpy.reshape(temp_b2_1,(p/m,p))
        gpu_multi.terminate()

#RESHAPE AGAIN TO (8192 * 8192) IMAGE
    for i in range(m):
        b2_a[i*p/m:(i+1)*p/m,:] = b2pa[:,:,i]

